#include "Record.h"
#include "MethodEntryRecord.h"
#include "MethodNameRecord.h"
#include "MethodExitRecord.h"
#include "AllocationRecord.h"
#include "PointerUpdateRecord.h"
#include "DeathRecord.h"
#include "FlatTraceOutputter.h"
#include "CallBackHandler.h"
#include "RootRecord.h"
#include <iostream>
#include <sstream>
#include <algorithm>
#include <vector>
#include <jvmti.h>
#include <agent_util.h>
#include <limits.h>
#include <assert.h>
#include <set>
#include "main.h"
using namespace std;

FlatTraceOutputter::FlatTraceOutputter(jvmtiEnv* jvmti, ostream *out, void (*fullcb)(void) )
  : output(out),
    // currentTime(0),
    aJVMTI(jvmti)
{
	initialize(fullcb, DEFAULT_MAX_RECORDS);
}

FlatTraceOutputter::FlatTraceOutputter(jvmtiEnv* jvmti, ostream *out, unsigned int recordLimit, void (*fullcb)(void))
  : output(out),
    // currentTime(0),
    aJVMTI(jvmti)
{
	initialize(fullcb, recordLimit);
}

/*Set up some monitors, zero some counters, allocate some buffers */
void FlatTraceOutputter::initialize(void (*fullcb)(void), unsigned int recordLimit){

	jvmtiError err;

	err = aJVMTI->CreateRawMonitor("yieldLock", &yieldLock);
	check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::FlatTraceOutputter Could not create yieldLock");

	err = aJVMTI->CreateRawMonitor("fullMon", &fullMon);
	check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::FlatTraceOutputter Could not create fullMon");

	err = aJVMTI->CreateRawMonitor("stdoutLock", &stdoutLock);
	check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::FlatTraceOutputter Could not create stdoutLock");


	//We need to keep some slack at the end of this buffer
	//Because we will copy death records onto the end of it for sorting
	realBufferSize = (unsigned int)(recordLimit * 1.5);
	deathBufferSize = recordLimit/2;

	assert(recordLimit + deathBufferSize == realBufferSize);

	outputBuffer = new Record * [realBufferSize];
	deathBuffer = new Record * [deathBufferSize];
	maxRecords = recordLimit;
	bufferEnd = 0;
	recFinishedCount = 0;
	deathBufferEnd = 0;
	gcsStarted = 0;
	gcsFinished = 0;
	
	// -- Counters
	countMethodEntries = 0;
	countMethodExits = 0;
	countAllocations = 0;
	countPointerUpdates = 0;
	countDeaths = 0;
	countVarEvents = 0;
	this->fullcb = fullcb;
}

FlatTraceOutputter::~FlatTraceOutputter()
{
}


void FlatTraceOutputter::doMethodEntry(unsigned long time, int methodId, /*string className, string methodName, string signature, */ unsigned long receiver, unsigned long thread)
{
	Record* entryRecord = NULL;

	entryRecord = new MethodEntryRecord(methodId, receiver, thread);
	entryRecord->setTime(time);
	addRecordToList(entryRecord);
	countMethodEntries++;

	//outputBuffer.push_back(entryRecord);
	
        // static long lastMethodID = 0;
	// string fullMethodName;

	// this->currentTime++;
	
	/*
	fullMethodName = className + methodName + signature;
	
	if(! seenBefore(fullMethodName)){
		Record* nameRecord = NULL;

		lockOn(method2idLock);
		method2id.insert( pair<string,long>(className+methodName+signature , ++lastMethodID ));
		unlockOn(method2idLock);

		nameRecord = new MethodNameRecord(lastMethodID, className, methodName, signature);
		
		nameRecord->setTime(currentTime);
		
		//outputBuffer.push_back(nameRecord);
		addRecordToList(nameRecord);
	}
	*/

	// entryRecord = new MethodEntryRecord(method2id[fullMethodName], receiver);
}

void FlatTraceOutputter::doMethodExit(unsigned long time, int methodId, /* string className, string methodName, string signature, */ unsigned long receiver, unsigned long thread)
{
	Record* exitRecord = NULL;
	// this->currentTime++;

	/*
	string fullMethodName = className + methodName + signature;

	unsigned long methodID = 0;
	if(! seenBefore(fullMethodName)){
	  // THIS IS BAD
	} else {
	  methodID = method2id[fullMethodName];
	}
	*/
	
	// exitRecord = new MethodExitRecord(methodID, receiver);
	exitRecord = new MethodExitRecord(methodId, receiver, thread);
	exitRecord->setTime(time);
	addRecordToList(exitRecord);
	countMethodExits++;
	//outputBuffer.push_back(exitRecord);
}

#ifdef THREAD_TAG_UPDATES
void FlatTraceOutputter::doPointerUpdate(unsigned long time, long oldTarget, long origin, long newTarget, long thread){
#else
void FlatTraceOutputter::doPointerUpdate(unsigned long time, long oldTarget, long origin, long newTarget){
#endif 
	Record* updateRecord = NULL;
	
	#ifdef THREAD_TAG_UPDATES
	updateRecord = new PointerUpdateRecord(oldTarget, origin, newTarget, thread);
	#else	
	updateRecord = new PointerUpdateRecord(oldTarget, origin, newTarget);
	#endif
	updateRecord->setTime(time);	
	addRecordToList(updateRecord);
	countPointerUpdates++;
}

void FlatTraceOutputter::doAllocation(unsigned long time, int objectTag, jlong size, string type, unsigned long thread)
{
	Record* allocationRecord= NULL;

	// unsigned long typeId = (unsigned long) type;
	allocationRecord = new AllocationRecord(objectTag, size, type, thread);
	allocationRecord->setTime(time);
	addRecordToList(allocationRecord);
	countAllocations++;
	//outputBuffer.push_back(allocationRecord);
}

void FlatTraceOutputter::doRoot(unsigned long time, unsigned long obj){
	Record* rootRecord = NULL;

	rootRecord = new RootRecord(obj,0);
	rootRecord->setTime(time);
	addRecordToList(rootRecord);

}

void FlatTraceOutputter::doVarEvent(unsigned long time, unsigned long obj, unsigned long thread)
{
  Record* rootRecord = NULL;

  rootRecord = new RootRecord(obj, thread);
  rootRecord->setTime(time);
  addRecordToList(rootRecord);

  countVarEvents++;
}


void FlatTraceOutputter::doDeath(long tag, long deathTime)
{
	Record* deathRecord = NULL;
	
	deathRecord = new DeathRecord(tag, deathTime);
	addRecordToDeathList(deathRecord);
	countDeaths++;
}

static bool recordCompare(const Record * a, const Record * b){
	return a->getTime() < b->getTime();		
}

void FlatTraceOutputter::doGCEnd(){
  // vector<Record*>::iterator it;

  stringstream verbose_string;

 
  verbose_string << "GCEND: Processing    " << "A " << countAllocations 
       << "   D " << countDeaths
       << "   U " << countPointerUpdates
       << "   M " << countMethodEntries
       << "   E " << countMethodExits 
       << "   V " << countVarEvents
       << endl;

  verbose_println(verbose_string.str());

  memcpy( &outputBuffer[recFinishedCount], deathBuffer, deathBufferEnd * sizeof(Record*));
  unsigned int combinedBufferSize =  recFinishedCount + deathBufferEnd;

  //Only read up to the finished records,
  //Do not read slots for records that haven't been inserted yet.

  stable_sort(&outputBuffer[0], &outputBuffer[combinedBufferSize], recordCompare);
	
  (*output) << hex;
  for (int i = 0; i < combinedBufferSize; i++) {
    outputBuffer[i]->streamTo(*output);
    // (*output) << endl;
    delete outputBuffer[i];
  }

  (*output) << dec;
  bufferEnd = 0;
  recFinishedCount = 0;
  deathBufferEnd = 0;

}

//Flush the trace output buffer. This is mostly for debugging purposes.
/*void FlatTraceOutputter::flush(){
  // vector<Record*>::iterator it;

  cout << "FLUSH: Processing " << bufferEnd << " records" << endl;

  memcpy( &outputBuffer[recFinishedCount], deathBuffer, deathBufferEnd * sizeof(Record*));
  //This is -2 because in the current state, bufferEnd is actually point _2_ past the end of the buffer
  unsigned int combinedBufferSize =  bufferEnd + deathBufferEnd -2;

  stable_sort(&outputBuffer[0], &outputBuffer[combinedBufferSize], recordCompare);


  for (int i = 0; i < combinedBufferSize; i++) {
    outputBuffer[i]->streamTo(*output);
    delete outputBuffer[i];
  }

  (*output) << dec;
  bufferEnd = 0;
  recFinishedCount = 0;
  deathBufferEnd = 0;
}*/

/*
void FlatTraceOutputter::addRecordToList(Record* rec, bool check_limit){
	jvmtiError error;

	int index = __sync_fetch_and_add(&bufferEnd, 1);
	if(bufferEnd > DEFAULT_MAX_RECORDS*2){
		cerr<< "FlatTraceOutputter::addRecordToList: Buffer exceeded." << endl;
		exit(-1);
	}

	outputBuffer[index] = rec;


	if (check_limit) {
	  if(bufferEnd >= maxRecords){
	    cout << "--GC Forced-- " << maxRecords  << endl;
	    (*fullcb)();
	  }
	}
}*/

void FlatTraceOutputter::addRecordToDeathList(Record* rec){
	int index = __sync_fetch_and_add(&deathBufferEnd,1);
	if(index >= deathBufferSize){
		cerr << "FlatTraceOutputter::addRecordToDeathList: Error, death buffer full.";
		exit(1);
	}

	deathBuffer[index] = rec;
	
}

void FlatTraceOutputter::addRecordToList(Record *rec){
	jvmtiError err;
	#ifdef ET_DEBUG
	jthread t;
	aJVMTI->GetCurrentThread(&t);
	int loops = 0;
	#endif

	while(true){	
		#ifdef ET_DEBUG
		loops++;
		#endif

		int index = __sync_fetch_and_add(&bufferEnd,1);
		if(index < maxRecords){
			outputBuffer[index] = rec;
			__sync_fetch_and_add(&recFinishedCount,1);

			#ifdef ET_DEBUG
			if (loops > 1) {
				aJVMTI->RawMonitorEnter(stdoutLock);
				cout << "(index < maxRecords): succeeded on try " << loops << " " << t << endl;
				aJVMTI->RawMonitorExit(stdoutLock);
			}
			#endif
			break;
		}
		else if(index == maxRecords){
			//Because we always get index via an atomic fetch and add,
			//There should be no possible way for two threads to get
			//into this case at the same time; thus this does not dead lock
			
			//Yield lock is basically a hack to make a JNI thread yield,
			//since we may need to wait for other threads to finish inserting 
			//their records.
			//Probably there is a better way to do this.
			err = aJVMTI->RawMonitorEnter(yieldLock);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not acquire yield lock");

			while(recFinishedCount < index){
				
				#ifdef ET_DEBUG
				aJVMTI->RawMonitorEnter(stdoutLock);
				cout << "(index == maxRecords): Waiting in loop..." << t << endl;
				aJVMTI->RawMonitorExit(stdoutLock);
				#endif

				err = aJVMTI->RawMonitorWait(yieldLock, 1);
				check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not wait on yield lock");
			}
		
			err = aJVMTI->RawMonitorExit(yieldLock);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter:addRecordToList: could not exit yieldlock");
			
			int gcnum = __sync_fetch_and_add(&gcsStarted, 1) + 1;

			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "(index == maxRecords): gc " << gcnum << ", loops=" << loops << ", trying to acquire fullMon. " << t << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif
			err = aJVMTI->RawMonitorEnter(fullMon);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter:: could not gain monitor ownership of fullMon");
			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "(index == maxRecords): got fullMon " << t << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif

			//Do the fullcb, this will trigger a GC.
			//We will get a bunch of death call backs, and a GCEnd event.
			//This is kind of non-modular, this module needs to know that this will cause a GCEvent.
			//Will refactor when time permits
			(*fullcb)();

			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "(index == maxRecords): back from fullcb; about to NotifyAll " << t << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif

			//When this returns, the GC is done, our buffers have been emptied.
			//Other threads waiting to add records to it can go ahead.
			__sync_fetch_and_add(&gcsFinished, 1);

			assert(gcsStarted == gcsFinished);

			err = aJVMTI->RawMonitorNotifyAll(fullMon);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not notify all listeners to fullMon");

			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "(index == maxRecords): about to exit fullMon " << t << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif 
			err = aJVMTI->RawMonitorExit(fullMon);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not exit fullMon");
			continue;
		}
		else{
			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "Else (index > maxRecords): Trying to enter fullmon " << t << " loops=" << loops << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif

			err = aJVMTI->RawMonitorEnter(fullMon);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not acqure fullMon (else clause)");

			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "Else (index > maxRecords: Full mon acquired " << t << endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif 

			while(gcsStarted != gcsFinished){
				#ifdef ET_DEBUG	
				aJVMTI->RawMonitorEnter(stdoutLock);
				cout << "Else (index > maxRecords): Waiting for fullMon... " << t <<  endl;
				aJVMTI->RawMonitorExit(stdoutLock);
				#endif
				
				err = aJVMTI->RawMonitorWait(fullMon,1000);
				check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not wait on fullMon");

				#ifdef ET_DEBUG
				aJVMTI->RawMonitorEnter(stdoutLock);
				cout << "Else (index > maxRecords): Back from wait, about to exit " << t<<  endl;
				aJVMTI->RawMonitorExit(stdoutLock);
				#endif
			}

		
			err = aJVMTI->RawMonitorExit(fullMon);
			check_jvmti_error(aJVMTI, err, "FlatTraceOutputter::addRecordToList: could not exit fullMon (else clause)");

			//try again
			#ifdef ET_DEBUG
			aJVMTI->RawMonitorEnter(stdoutLock);
			cout << "Else (index > maxRecords): about to try again " << t<<  endl;
			aJVMTI->RawMonitorExit(stdoutLock);
			#endif
			continue;		
		}
	}
}


