package classReWriter;

import org.objectweb.asm.*;
import org.objectweb.asm.commons.*;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

public class ETOptMethodAdapter extends MethodAdapter implements Opcodes {
	
  private static final boolean wrapAll = true;
  private static int wrapOnly = 184; // 184
  private static int seenSoFar = 0;
  private static final String nowrapClass = "java/lang/ClassLoaderxxx";
  private static final String nowrapMethod = "loadClass";
  private static final String nowrapDesc = "(Ljava/lang/String;Z)Ljava/lang/Class;";

  /**
   * indicate whether we have seen an instruction yet
   */
  private boolean beforeFirstInsn = true;

  /**
   * indicates if we are in a method safe to instrument;
   * avoids inserting code in a constructor before the call
   * to the super's constructor, so as not to break JVM rules;
   * equivalently, indicates that onMethodEnter has been called
   */
  private boolean insideMethod = false;

  /**
   * indicates whether to timestamp arguments and this on entry
   */
  private static final boolean STACK_SCAN_ON_ENTRY = true; 
	
  /**
   * name of the current method
   */
  protected final String methodName;

  /**
   * type descriptor of the current method
   */
  protected final String methodDesc;

  /**
   * name of the current class
   */
  protected final String className;

  /**
   * access (public, protected, etc.) of the current method
   */
  protected final int access;

  /**
   * the particular experimental features enabled
   */
  protected final Set<String> expFlags;

  /**
   * unique id that we have assigned to the current method
   */
  protected final int methodId;
	
  /**
   * maintains value numbers for the local variables
   */
  protected IntStack locals = new IntStack();

  /**
   * maintains value numbers for the working stack
   */
  protected IntStack stack = new IntStack();

  /**
   * next value number to use
   */
  protected int vnCounter = 0;

  /**
   * values timestamped since last reset
   */
  protected Set<Integer> stamped = new HashSet<Integer>();

  /**
   * a map for holding numbers we assigned to uninitialized (new) objects
   */
  private Map<Object, Integer> uninits = new HashMap<Object, Integer>();

  /**
   * a counter for assigning numbers to uninitialized objects, based
   * on the label at which they were allocated
   */
  private int uninitCounter = -1;

  /**
   * true iff we are processing an instance constructor (init method)
   */
  private boolean inConstructor;

  /**
   * turn off to avoid instrumenting a method
   */
  private boolean instrument = true;

  /**
   * maps the index of an uninit object (as seen in the stack model)
   * to the unique id of the site that generated it;
   * -1 always maps to 0, meaning "uninitialized this"
   */
  protected final Map<Integer, Integer> newSiteId = new HashMap<Integer, Integer>();

  /**
   * maps a Label at a NEW to the NEW's site number;
   * not every NEW has a Label right on it, but those NEWs
   * are not referred to by Frames so we will never look them up
   */
  protected final Map<Object, Integer> siteForLabel = new HashMap<Object, Integer>();

  /**
   * last new site id assigned
   */
  private static int lastNewSiteId = 0;

  /**
   * label for start of the method, if we are adding a wrapped exception handler
   */
  private Label startLabel;

  /**
   * label for end of the (original) method, if we are adding a wrapped exception handler
   */
  private Label endLabel;

  /**
   * most recent label seen
   */
  private Label lastLabel;

  /**
   * checks whether a particular experimental flag is set;
   * we use this help method incase we want a richer data structure
   * later, such as a Map instead of a Set
   * @param flag a String giving the flag to check
   * @result a boolean, true iff the flag is set
   */
  protected boolean haveXFlag (String flag) {
    return expFlags.contains(flag);
  }

  /**
   * constructor to build an adapter for a particular class;
   * initializes fields, assigns a new unique id, and outputs the uid info
   */
  public ETOptMethodAdapter (int access, String className,
                             String name, String desc, MethodVisitor mv,
                             Set<String> expFlags, PrintStream nameStream) {
    super(mv);
    // super(mv, access, name, desc);
    this.className    = className;
    this.methodName   = name;
    this.methodDesc   = desc;
    this.access       = access;
    this.expFlags     = expFlags;
    this.methodId     = ClassReWriter.getNextMethodId();

    int classId = ETAdapter.getClassId(className);

    nameStream.printf("N 0x%x 0x%x %s %s %s%n",
                      methodId, classId, className, methodName, methodDesc);

    inConstructor = name.equals("<init>");
  }

  /**
   * Looks through the descriptor to determine which parameter local var slots
   * contain object references; useful for timestamping them, for example
   * @param ignoreThis a boolean that when true means to skip "this"
   */
  protected Vector<Integer> findParamSlots (boolean ignoreThis) {
    Vector<Integer> slots = new Vector<Integer>();
    int slotIndex = 0;  // next local variable slot

    if (((access & Opcodes.ACC_STATIC) == 0)) {
      if (!ignoreThis) {
        // deal with 'this' for non-static methods
        slots.add(slotIndex); // timestamp here to show live after method entry
      }
      slotIndex++;
    }

    if (methodDesc.charAt(0) != '(') {
      // throw some exception?
      return slots;
    }
    // start i at 1 to skip the left-paren '('
    scan: for (int i = 1; ; i++) { // note: runs to the right paren
      switch (methodDesc.charAt(i)) {
      case ')':
        break scan;

      case 'D': case 'J':
        slotIndex += 2;
        continue;

      case 'L':
        slots.add(slotIndex++);
        while (methodDesc.charAt(i) != ';') { i++; }
        continue;

      case '[':
        slots.add(slotIndex++);
        // skip over all dimensions
        while (methodDesc.charAt(i) == '[') { i++; }
        // if an array of bject type, skip over the object type
        if (methodDesc.charAt(i) == 'L') {
          while (methodDesc.charAt(i) != ';') { i++; }
        }
        continue;

      default:
        slotIndex += 1;
        continue;
      }
    }
    return slots;
  }

  /**
   * Based on the given descriptor, does the given method return an object/array?
   */
  public boolean returnsObject (String desc){
    int closeIdx = desc.indexOf(')');
    char retKind = desc.charAt(closeIdx + 1);
    return (retKind == 'L' || retKind == '[');
  }

  /**
   * get the unique id of the allocating site for a value number
   * that decribes an uninitialized object; returns 0 for uninitialized this
   * @param vn an int giving the value number, which must be for an uninit object
   * @return an int giving the unique allocating site, or 0 for "uninitialized this"
   */
  protected int getNewSiteId (int vn) {
    Integer vni = vn;
    if (!newSiteId.containsKey(vni)) {
      // return -100;
      throw new IllegalArgumentException("arg was " + vn);
    }
    return newSiteId.get(vni);
  }

  /**
   * Instruments Object.<init> to note the new object;
   * also controls whether to instrument a class at all,
   * and thus avoids instrumenting Object.finalize, which
   * would lead to problems.
   */
  @Override
  public void visitCode () {

    boolean inObject = className.equals("java/lang/Object");

    if (inObject && methodName.equals("finalize")) {
      // don't instrument finalize() -- causes infinite recursion!
      instrument = false;
    }

    if (instrument && haveXFlag("W")) {
      // Code for putting a handler around the method
      // Defer emitting the try block until just before the first instruction;
      //   this insures that other handlers starting at 0 come first, and our
      //   new one match only *after* them (important for correctness since the
      //   JVM run-time takes the first applicable handler)
      // Note that this means we must override some visit methods just for
      //   checking for the first instruction; however, we need not do so
      //   for instructions that examine the operand stack since the stack
      //   is empty at the first instruction.  See checkFirstInsn and its callers.
      startLabel = new Label();
      endLabel   = new Label();
    }

    if (inObject && methodName.equals("<init>") && haveXFlag("O")) {
      // Note new object
      mv.visitVarInsn(Opcodes.ALOAD, 0);
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newobj",
                         "(Ljava/lang/Object;)V");
      // continue
    }

    super.visitCode();

    if (haveXFlag("E")) {
      onMethodEnter(inConstructor);
    } else if (!inConstructor) {
      onMethodEnter(false);
    }

    // initialize locals
    if ((access & Opcodes.ACC_STATIC) == 0) {
      // deal with 'this' for non-static methods
      if (inConstructor) {
        locals.push(-1);
        newSiteId.put(-1, 0);
      } else {
        locals.push(++vnCounter);
        nowStamped(vnCounter);
      }
    }

    // start i at 1 to skip the left-paren '('
    scan: for (int i = 1; ; i++) { // note: runs to the right paren
      switch (methodDesc.charAt(i)) {
      case ')':
        break scan;

      case 'D': case 'J':
        locals.pushZero(2);
        continue;

      case 'L':
        locals.push(++vnCounter);
        stampedIf(vnCounter, !inConstructor);
        while (methodDesc.charAt(i) != ';') { i++; }
        continue;

      case '[':
        locals.push(++vnCounter);
        stampedIf(vnCounter, !inConstructor);
        // skip over all dimensions
        while (methodDesc.charAt(i) == '[') { i++; }
        // if an array of bject type, skip over the object type
        if (methodDesc.charAt(i) == 'L') {
          while (methodDesc.charAt(i) != ';') { i++; }
        }
        continue;

      default:
        locals.push(0);
        continue;
      }
    }

  }


  /**
   * processes a Frame to get ready for the next (extended) basic block
   */
  @Override
  public void visitFrame (int type, int nLocal, Object[] locals, int nStack, Object[] stack) {
    if (type != Opcodes.F_NEW) { // uncompressed frame
      throw new
        IllegalStateException("ClassReader.accept() should be called with EXPAND_FRAMES flag");
    }
    vnCounter = 0;
    uninitCounter = -1;
    uninits.clear();
    visitFrameInfo(this.locals, nLocal, locals);
    visitFrameInfo(this.stack , nStack, stack );
    super.visitFrame(type, nLocal, locals, nStack, stack);
    tick();  // counts as a tick since we don't know anything
  }

  /**
   * helper subroutine to initialize an IntStack with value numbers
   * according the types given (either locals or working stack)
   * @param s the IntStack to set up; this routine clears it first
   * @param n an int giving the number of items to set up (as for Frames)
   * @param infos an Object array with the information; must have size at least n
   */
  private void visitFrameInfo (IntStack s, final int n, Object[] infos) {
    s.clear();
    for (int i = 0; i < n; ++i) {
      Object info = infos[i];
      if (info == Opcodes.UNINITIALIZED_THIS) {
        // uninitialized "this"
        s.push(-1);
      } else if (info instanceof String) {
        // an Object type
        s.push(++vnCounter);
      } else if (info instanceof Label) {
        // an uninitialized object (other than "this")
        if (uninits.containsKey(info)) {
          s.push(uninits.get(info));
        } else {
          s.push(--uninitCounter);
          uninits.put(info, uninitCounter);
          Integer siteInt = siteForLabel.get(info);
          newSiteId.put(uninitCounter, siteInt);
          /*
          System.err.printf("Frame: method = %x, local # = %d, site = %d, Label = %s%n",
                            methodId, uninitCounter, siteInt, info);
          */
        }
      } else {
        // a primitive, NULL, or TOP
        s.push (0);
        if (info == Opcodes.LONG || info == Opcodes.DOUBLE) {
          // a double length primitive
          s.push (0);
        }
      }
    }
  }

  @Override
  public void visitInsn (int opcode) {
    checkFirstInsn();

    // first check for onMethodExit work
    switch (opcode) {
    case IRETURN:
    case LRETURN:
    case FRETURN:
    case DRETURN:
    case ARETURN:
    case RETURN:
      onMethodExit(opcode);
      break;

    case ATHROW:
      if (haveXFlag("T")) {
        onMethodExit(opcode);
      }
      break;
    }

    outer: {
      switch (opcode) {
        // Note: cases that "break outer" have done the work of the original byte code

      case AALOAD:
      case BALOAD:
        if (shouldStamp(1)) {
          // use for BALOAD since combines byte and boolean, for AALOAD because of return type
          mv.visitInsn(Opcodes.DUP2);
          // Stack is a,idx,a,idx
          mv.visitInsn(Opcodes.POP);
          // Stack is a,idx,a
          stampOne();
        }
        break;

      case CALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "CALoadEvent", "([CI)C");
          break outer;
        }
        break;

      case FALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "FALoadEvent", "([FI)F");
          break outer;
        }
        break;

      case IALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "IALoadEvent", "([II)I");
          break outer;
        }
        break;

      case SALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "SALoadEvent", "([SI)S");
          break outer;
        }
        break;

      case LALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "LALoadEvent", "([JI)J");
          break outer;
        }
        break;

      case DALOAD:
        if (shouldStamp(1)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "DALoadEvent", "([DI)D");
          break outer;
        }
        break;

      case AASTORE:
        // always instrument this: creates a heap edge; never an uninit
        if (instrument) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "AAStoreEvent",
                             "([Ljava/lang/Object;ILjava/lang/Object;)V");
        }
        break outer;

      case CASTORE:
        if (shouldStamp(2)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "CAStoreEvent", "([CIC)V");
          break outer;
        }
        break;

      case FASTORE:
        if (shouldStamp(2)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "FAStoreEvent", "([FIF)V");
          break outer;
        }
        break;

      case IASTORE:
        if (shouldStamp(2)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "IAStoreEvent", "([III)V");
          break outer;
        }
        break;

      case SASTORE:
        if (shouldStamp(2)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "SAStoreEvent", "([SII)V");
          break outer;
        }
        break;

      case BASTORE:
        if (shouldStamp(2)) {
          mv.visitInsn(Opcodes.DUP2_X1);  // idx, val, arrayref, idx, val
          mv.visitInsn(Opcodes.POP2);     // idx, val, arrayref
          mv.visitInsn(Opcodes.DUP_X2);   // arrayref, idx, val, arrayref
          stampOne();
        }
        break;

      case DASTORE:
        if (shouldStamp(3)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "DAStoreEvent", "([DID)V");
          break outer;
        }
        break;

      case LASTORE:
        if (shouldStamp(3)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "LAStoreEvent", "([JIJ)V");
          break outer;
        }
        break;

      case ARRAYLENGTH:
      case ATHROW:
      case MONITORENTER:
      case MONITOREXIT:
        if (shouldStamp(0)) {
          stampDup();
        }
        break;

      }
      super.visitInsn(opcode);	
    } // end of outer

    // optimizer processing
    switch (opcode) {
    case NOP:
    case INEG:
    case LNEG:
    case FNEG:
    case DNEG:
    case I2F:
    case L2D:
    case F2I:
    case D2L:
    case I2B:
    case I2C:
    case I2S:
      // no net stack change
    case IRETURN:
    case LRETURN:
    case FRETURN:
    case DRETURN:
    case ARETURN:
    case RETURN:
    case ATHROW:
      // these will terminate a block, so don't bother
      break;

    case ACONST_NULL:
    case ICONST_M1:
    case ICONST_0:
    case ICONST_1:
    case ICONST_2:
    case ICONST_3:
    case ICONST_4:
    case ICONST_5:
    case FCONST_0:
    case FCONST_1:
    case FCONST_2:
    case I2L:
    case I2D:
    case F2L:
    case F2D:
      // push one primtive
      stack.push(0);
      break;

    case LCONST_0:
    case LCONST_1:
    case DCONST_0:
    case DCONST_1:
      // push two primitive slots
      stack.pushZero(2);
      break;

    case IALOAD:
    case FALOAD:
    case BALOAD:
    case CALOAD:
    case SALOAD:
      // pop array, idx, push prim
      stack.pop();
      nowStamped(stack.pop());
      stack.push(0);
      break;

    case LALOAD:
    case DALOAD:
      // pop array, idx, push two prim slots
      stack.pop();
      nowStamped(stack.pop());
      stack.pushZero(2);
      break;

    case AALOAD:
      // pop array, idx; push new object
      stack.pop();
      nowStamped(stack.pop());
      nowStamped(stack.push(++vnCounter));
      break;

    case IASTORE:
    case FASTORE:
    case BASTORE:
    case CASTORE:
    case SASTORE:
      // pop array, idx, value
      stack.pop(2);
      nowStamped(stack.pop());
      break;

    case AASTORE:
      // pop array, idx, value
      nowStamped(stack.pop());
      stack.pop();
      nowStamped(stack.pop());
      break;

    case LCMP:
    case DCMPL:
    case DCMPG:
      // pop 4 prims, push one prim
      stack.pop(3);
      break;

    case LASTORE:
    case DASTORE:
      // pop array, idx, two prims
      stack.pop(3);
      nowStamped(stack.pop());
      break;

    case MONITORENTER:
    case MONITOREXIT:
      // pop one slot
      nowStamped(stack.pop());
      break;
      
    case POP:
      // pop one slot
    case IADD:
    case FADD:
    case ISUB:
    case FSUB:
    case IMUL:
    case FMUL:
    case IDIV:
    case FDIV:
    case IREM:
    case FREM:
    case ISHL:
    case ISHR:
    case IUSHR:
    case IAND:
    case IOR:
    case IXOR:
    case L2I:
    case L2F:
    case D2I:
    case D2F:
    case FCMPL:
    case FCMPG:
      // pop two prims, push one
    case LSHL:
    case LSHR:
    case LUSHR:
      // pop 3 prims, push 2
      stack.pop();
      break;

    case POP2:
      // pop two slots
    case LADD:
    case DADD:
    case LSUB:
    case DSUB:
    case LMUL:
    case DMUL:
    case LDIV:
    case DDIV:
    case LREM:
    case DREM:
    case LAND:
    case LOR:
    case LXOR:
      // pop 4 prims, push 2
      stack.pop(2);
      break;

    case DUP:
      stack.push(stack.top());
      break;

    case DUP_X1: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      stack.push(v1);
      stack.push(v2);
      stack.push(v1);
      break;
    }
        
    case DUP_X2: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      int v3 = stack.pop();
      stack.push(v1);
      stack.push(v3);
      stack.push(v2);
      stack.push(v1);
      break;
    }

    case DUP2: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      stack.push(v2);
      stack.push(v1);
      stack.push(v2);
      stack.push(v1);
      break;
    }

    case DUP2_X1: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      int v3 = stack.pop();
      stack.push(v2);
      stack.push(v1);
      stack.push(v3);
      stack.push(v2);
      stack.push(v1);
      break;
    }

    case DUP2_X2: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      int v3 = stack.pop();
      int v4 = stack.pop();
      stack.push(v2);
      stack.push(v1);
      stack.push(v4);
      stack.push(v3);
      stack.push(v2);
      stack.push(v1);
      break;
    }

    case SWAP: {
      int v1 = stack.pop();
      int v2 = stack.pop();
      stack.push(v1);
      stack.push(v2);
      break;
    }

    case ARRAYLENGTH:
      nowStamped(stack.pop());
      stack.push(0);
      break;

    }

  }
  

  /**
   * handles field insns; timestamps accesses to reference fields
   */
  @Override
  public void visitFieldInsn (int opcode, String owner, String fieldName, String fieldDesc) {
    checkFirstInsn();

    int w = width(fieldDesc);
    boolean isRef = isObjectReference(fieldDesc);

    switch (opcode) {

    case PUTFIELD:
      if (isRef) {
        // storing a pointer
        if (stack.top(1) < 0) {
          // storing into an uninit obj (which is legal)
          if (haveXFlag("U") && instrument) {
            // new version indicates update of an uninit object
            mv.visitInsn(Opcodes.DUP); // ...,o,v,v
            int newId = getNewSiteId(stack.top(1));
            mv.visitLdcInsn(newId);    // ...,o,v,v,o' (o' is an id for o)
            int fieldId = (true ? 0 : ETAdapter.getFieldId(owner, fieldName, false));
            mv.visitLdcInsn(fieldId);  // ...,o,v,v,o',f
            /*
            System.err.printf("Uninit PUTFIELD: method = %x, local # = %d, site = %d, field = %d%n",
                              methodId, stack.top(1), newId, fieldId);
            */
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "uninitPutfield",
                               "(Ljava/lang/Object;II)V");
          } else {
            // can't call pointerMovedOff;
            // just stamp the object being stored
            if (shouldStamp(0)) {
              stampDup();
            }
          }
        } else {
          // storing a pointer into a regular (initialized) object
          if (instrument) {
            if (haveXFlag("F")) {
              // Reference case: call pointerMovedOff
              // Stack is ...,o,v
              mv.visitInsn(Opcodes.DUP2); // ...,o,v,o,v
              mv.visitInsn(Opcodes.SWAP); // ...,o,v,v,o
              mv.visitInsn(Opcodes.DUP);  // ...,o,v,v,o,o
              mv.visitFieldInsn(Opcodes.GETFIELD, owner, fieldName, fieldDesc);
              // ...,o,v,v,o,old
              mv.visitLdcInsn((true ? 0 : ETAdapter.getFieldId(owner, fieldName, false)));
              mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff",
                                 "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;I)V");
              // TODO: rewrite the above to:
              // DUP2, push fieldId, call pointerUpdated
            } else {
              // Reference case: call pointerMovedOff
              // Stack is ...,o,v
              mv.visitInsn(Opcodes.DUP2); // ...,o,v,o,v
              mv.visitInsn(Opcodes.SWAP); // ...,o,v,v,o
              mv.visitInsn(Opcodes.DUP);  // ...,o,v,v,o,o
              mv.visitFieldInsn(Opcodes.GETFIELD, owner, fieldName, fieldDesc);
              // ...,o,v,v,o,old
              mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff",
                                 "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
            }
            // ...,o,v
          }
        }
      } else if (w == 2) {
        // a long or double field
        if (shouldStamp(2)) {
          // wide field case: call varEvent1
          // ...,o,wide
          mv.visitInsn(Opcodes.DUP2_X1);  // ...,wide,o,wide
          mv.visitInsn(Opcodes.POP2);     // ...,wide,o
          mv.visitInsn(Opcodes.DUP_X2);   // ...,o,wide,o
          stampOne();
        }
      } else {
        // a length-one primitive field
        if (shouldStamp(1)) {
          // narrow field case: call varEvent1
          // ...,o,v
          mv.visitInsn(Opcodes.DUP2);     // ...,o,v,o,v
          mv.visitInsn(Opcodes.POP);      // ...,o,v,o
          stampOne();
        }
      }
      break;
		  
    case Opcodes.PUTSTATIC:
      if (isRef && instrument) {
        if (haveXFlag("F")) {
          // Stack is ...,v;
          mv.visitInsn(Opcodes.DUP); // ...,v,v;
          mv.visitFieldInsn(Opcodes.GETSTATIC, owner, fieldName, fieldDesc);
          // Stack is ...,v,v,oldVal
          mv.visitLdcInsn(true ? 0 : ETAdapter.getFieldId(owner, fieldName, true));
          // Stack is ...,v,v,oldVal,fieldId
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff",
                             "(Ljava/lang/Object;Ljava/lang/Object;I)V");
          // TODO: rewrite to DUP, push fieldId, call pointerUpdated
          // Stack is ...,v;
        } else {
          // Stack is ...,v;
          mv.visitInsn(Opcodes.DUP); // ...,v,v;
          mv.visitFieldInsn(Opcodes.GETSTATIC, owner, fieldName, fieldDesc);
          // Stack is ...,v,v,oldVal
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff",
                             "(Ljava/lang/Object;Ljava/lang/Object;)V");
          // Stack is ...,v;
        }
      }
      break;
          
    case Opcodes.GETFIELD:
      if (shouldStamp(0)) {
        stampDup();
      }
      break;

    }

    super.visitFieldInsn(opcode, owner, fieldName, fieldDesc);

    switch (opcode) {

    case GETFIELD:
      nowStamped(stack.pop());
      // fall through
    case GETSTATIC:
      if (isRef) {
        stack.push(++vnCounter);
      } else {
        stack.pushZero(w);
      }
      break;

    case PUTFIELD:
      nowStamped(stack.top());  // works because nowStamped filters its argument
      stack.pop(w);
      nowStamped(stack.pop());
      break;

    case PUTSTATIC:
      nowStamped(stack.top());  // works because nowStamped filters its argument
      stack.pop(w);
      break;

    }

  }
		
  @Override
  public void visitJumpInsn (int opcode, Label l) {
    checkFirstInsn();

    switch (opcode) {
    case IF_ACMPEQ:
    case IF_ACMPNE:
      if (shouldStamp(1) || shouldStamp(0)) {
        // cannot have uninits here, so this test is ok
        mv.visitInsn(Opcodes.DUP2);
        mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent2",
                           "(Ljava/lang/Object;Ljava/lang/Object;)V");
      }
      break;

    case IFNULL:
    case IFNONNULL:
      if (shouldStamp(0)) {
        stampDup();
      }
      break;
      
    }

    super.visitJumpInsn(opcode,l);

    switch (opcode) {
    case IFEQ:
    case IFNE:
    case IFLT:
    case IFGE:
    case IFGT:
    case IFLE:
      stack.pop();
      break;

    case IF_ICMPEQ:
    case IF_ICMPNE:
    case IF_ICMPLT:
    case IF_ICMPGE:
    case IF_ICMPGT:
    case IF_ICMPLE:
      stack.pop(2);
      break;

    case IF_ACMPEQ:
    case IF_ACMPNE:
      nowStamped(stack.pop());
      nowStamped(stack.pop());
      break;

    case IFNULL:
    case IFNONNULL:
      nowStamped(stack.pop());
      break;

    case GOTO:
    case JSR:
      // ignore: ends a chunk of interest
      break;
    }
  }

  @Override
  public void visitIntInsn (int opcode, int operand) {
    checkFirstInsn();

    super.visitIntInsn(opcode, operand);

    if (opcode == Opcodes.NEWARRAY && instrument) {
      mv.visitInsn(Opcodes.DUP);
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newarray",
                         "(Ljava/lang/Object;)V");
    }

    switch (opcode) {

    case BIPUSH:
    case SIPUSH:
      stack.push(0);
      break;

    case NEWARRAY:
      stack.pop();
      nowStamped(stack.push(++vnCounter));
      break;

    }

  }

  @Override
  public void visitLdcInsn (Object obj) {
    checkFirstInsn();

    super.visitLdcInsn(obj);
    if (obj instanceof String || obj instanceof Type) {
      stack.push(++vnCounter);
    } else if (obj instanceof Long || obj instanceof Double) {
      stack.pushZero(2);
    } else {
      stack.push(0);
    }
  }

  @Override
  public void visitMethodInsn (int opcode, String owner, String name, String desc) {
    checkFirstInsn();

    // System.err.println("MethodInsn opcode=" + opcode + " callee=" + owner +
    //                    "." + name + desc);

    // pop arguments off stack
    int idx = 1; // skip the '('
    outer:
    while (true) {
      switch (desc.charAt(idx++)) {
      case 'B':
      case 'C':
      case 'F':
      case 'I':
      case 'S':
      case 'Z':
        stack.pop();
        continue;

      case 'D':
      case 'J':
        stack.pop(2);
        continue;

      case 'L':
        stack.pop();
        while (desc.charAt(idx++) != ';');
        continue;

      case '[':
        stack.pop();
        while (desc.charAt(idx) == '[')
          ++idx;
        if (desc.charAt(idx++) == 'L') {
          while (desc.charAt(idx++) != ';');
        }
        continue;

      case ')':
        break outer;
      }
    }

    if (opcode == Opcodes.INVOKESPECIAL && stack.top() < 0 && name.equals("<init>")) {
      if (instrument && haveXFlag("I")) {
        int siteId = newSiteId.get(stack.top());
        mv.visitLdcInsn(siteId);
        /*
        System.err.printf("Init: method = %x, local # = %d, site = %d%n",
                          methodId, stack.top(), siteId);
        */
        mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "initOf", "(I)V");
      }
    }

    super.visitMethodInsn(opcode, owner, name, desc);

    if (opcode == Opcodes.INVOKESPECIAL && stack.top() < 0 && name.equals("<init>")) {
      // found an initializer call!
      int oldTop = stack.top();
      foundInit(stack.top());
      if (oldTop == -1 && haveXFlag("L") && (!insideMethod)) {
        onMethodEnter(false);
      }
      if (haveXFlag("R") &&
          owner.equals("java/lang/ref/Reference") && name.equals("<init>") &&
          ETAdapter.isReferenceSubclass(className) && inConstructor) {
        // Here we are:
        // 1) in a constructor of an immediate subclass of java/lang/ref/Reference
        // 2) calling a constructor of java/lang/ref/Reference
        mv.visitVarInsn(Opcodes.ALOAD, 0);  // Reference object
        mv.visitVarInsn(Opcodes.ALOAD, 1);  // referent object
        mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "refInit",
                           "(Ljava/lang/ref/Reference;Ljava/lang/Object;)V");
      }
    }

    if (opcode != Opcodes.INVOKESTATIC) {
      stack.pop();
    }

    tick();

    switch (desc.charAt(idx)) {
    case 'B':
    case 'C':
    case 'F':
    case 'I':
    case 'S':
    case 'Z':
      stack.push(0);
      break;

    case 'D':
    case 'J':
      stack.pushZero(2);
      break;

    case 'L':
    case '[':
      nowStamped(stack.push(++vnCounter));
      break;

    }

    if (returnsObject(desc)) { // the method returns an object: timestamp it now
      stampDup();
    }	
  }

  @Override
  public void visitMultiANewArrayInsn (String desc, int dims) {
    checkFirstInsn();

    super.visitMultiANewArrayInsn(desc, dims);

    if (instrument) {
      mv.visitInsn(Opcodes.DUP);
      // This requires that dims actually be a short.
      // I hope no one wants an array with more than 32,767 dimensions
      mv.visitIntInsn(Opcodes.SIPUSH, dims);
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "multiNewArray",
                         "([Ljava/lang/Object;I)V");
    }
    stack.pop(dims);
    nowStamped(stack.push(++vnCounter));
  }

  @Override
  public void visitTypeInsn (int opcode, String type) {
    checkFirstInsn();

    outer: {
      switch (opcode) {

      case ANEWARRAY:
        super.visitTypeInsn(opcode,type);
        if (instrument) {
          mv.visitInsn(Opcodes.DUP);
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newarray",
                             "(Ljava/lang/Object;)V");
        }
        break outer;
          
      case CHECKCAST:
      case INSTANCEOF:
        if (shouldStamp(0)) {
          stampDup();
        }
        break;
          
      }
      super.visitTypeInsn(opcode, type);
    }
    
    switch (opcode) {

    case NEW:
      stack.push(--uninitCounter);
      if (haveXFlag("N")) {
        mv.visitLdcInsn((true ? 0 : ETAdapter.getClassId(type)));
        newSiteId.put(uninitCounter, ++lastNewSiteId);
        mv.visitLdcInsn(lastNewSiteId);
        if (ETAdapter.isReferenceInheritor(type)) {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newrefobj", "(II)V");
        } else {
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newobj", "(II)V");
        }
        siteForLabel.put(lastLabel, lastNewSiteId);
        /*
        System.err.printf("NEW: method = %x, local # = %d, site = %d, Label = %s%n",
                          methodId, uninitCounter, lastNewSiteId,
                          (lastLabel == null ? "null" : lastLabel.toString()));
        */
      }
      break;

    case ANEWARRAY:
      stack.pop();
      nowStamped(stack.push(++vnCounter));
      break;

    case CHECKCAST:
      nowStamped(stack.top());
      break;

    case INSTANCEOF:
      nowStamped(stack.pop());
      stack.push(0);
      break;
    }
  }

  @Override
  public void visitVarInsn (int opcode, int var) {
    checkFirstInsn();

    super.visitVarInsn(opcode, var);

    switch (opcode) {
    case ILOAD:
    case FLOAD:
      stack.push(0);
      break;

    case ALOAD:
      stack.push(locals.get(var));
      break;

    case LLOAD:
    case DLOAD:
      stack.pushZero(2);
      break;

    case ISTORE:
    case FSTORE:
      stack.pop();
      locals.set(var, 0);
      break;

    case LSTORE:
    case DSTORE:
      stack.pop(2);
      locals.set(var  , 0);
      locals.set(var+1, 0);
      break;

    case ASTORE:
      locals.set(var, stack.pop());
      break;

    case RET:
      break;
    }
  }
		
  @Override
  public void visitIincInsn (int var, int incr) {
    checkFirstInsn();

    super.visitIincInsn(var, incr);
  }

  /**
   * called when we can instrument entry to a method
   * For most methods this is immediately, but for constructors
   * we may have to wait until the super-constructor call.
   */
  protected void onMethodEnter (boolean inConstructor) {
    // System.err.println("onMethodEnter: " + className + "::" + methodName);

    insideMethod = true;

    if (!instrument) {
      return;
    }

    if (STACK_SCAN_ON_ENTRY) {
      // -- Emit method ID
      mv.visitLdcInsn(methodId);
					
      if ((access & Opcodes.ACC_STATIC) != 0) {
	// System.err.println("ADD-ENTRY: " + className + "::" + methodName);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodEntryStatic", 
			   "(I)V" );
      } else if (inConstructor) {
	mv.visitInsn(Opcodes.ACONST_NULL);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodEntry", 
			   "(ILjava/lang/Object;)V");
      } else {
	// System.err.println("ADD-ENTRY: " + className + "::" + methodName);
	mv.visitVarInsn(Opcodes.ALOAD, 0);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodEntry", 
			   "(ILjava/lang/Object;)V");
      }
    }

    Vector<Integer> paramSlots = findParamSlots(inConstructor);

    // load all the parameters to the operand stack
    
    for (int ps: paramSlots) {
      mv.visitVarInsn(Opcodes.ALOAD, ps);
    }
    
    // In case there are a lot of parameters,
    // stamp them using as few calls to varEvent 5 as possible.
    for (int i = 0; i < paramSlots.size()/5; i++) {
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent5",
                         "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;" +
                         "Ljava/lang/Object;Ljava/lang/Object;)V");
    }
    
    // Then get the remainder
    switch (paramSlots.size() % 5) {
    case 0:
      break;
    case 1:
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1",
                         "(Ljava/lang/Object;)V");
      break;
    case 2:
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent2",
                         "(Ljava/lang/Object;Ljava/lang/Object;)V");
      break;
    case 3:
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent3",
                         "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
      break;
    case 4:
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent4",
                         "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;" +
                         "Ljava/lang/Object;)V");
      break;
    }

  }

  /**
   * called when we are about to return or throw
   * @param opcode an int giving the opcode of the instruction causing the exit
   */
  protected void onMethodExit (int opcode) {
    // System.err.println("onMethodExit: " + className + "::" + methodName);
    // if( (opcode >= Opcodes.IRETURN && opcode <= Opcodes.RETURN) || opcode == Opcodes.ATHROW ){
      
    if (!instrument) {
      return;
    }

    // -- Emit method ID
    mv.visitLdcInsn(methodId);
      
    if ((access & Opcodes.ACC_STATIC) != 0) {
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExitStatic", "(I)V");
    } else {
      mv.visitVarInsn(Opcodes.ALOAD, 0);
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExit", 
                         "(ILjava/lang/Object;)V");
    }
  }

  @Override
  public void visitLabel (Label l) {
    // ClassReader guarantees one Label per instruction
    // The only possible extras are ones we add, and they (presently)
    //   do not interfere with the logic of this method
    super.visitLabel(l);
    lastLabel = l;
    /*
    System.err.printf("Label: method = %x, %s%n", methodId, lastLabel.toString());
    */
  }

  @Override
  public void visitMaxs (int maxStack, int maxLocals) {
    // complete the wrapped handler (if any)
    if (endLabel != null) {
      mv.visitLabel(endLabel);

      mv.visitInsn(Opcodes.DUP);
      mv.visitLdcInsn(methodId);

      if ((access & Opcodes.ACC_STATIC) != 0) {
        mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExceptionStatic", "(Ljava/lang/Object;I)V");
      } else {
        if (inConstructor) {
          mv.visitInsn(ACONST_NULL);
        } else {
          mv.visitVarInsn(Opcodes.ALOAD, 0);
        }
        mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodException",
                           "(Ljava/lang/Object;ILjava/lang/Object;)V");
      }
      mv.visitInsn(Opcodes.ATHROW);
    }
    super.visitMaxs(maxStack, maxLocals);
  }

  /**
   * checks if we are just before the first instruction, and if so,
   * does actions needed right before the instruction
   */
  protected void checkFirstInsn () {
    if (beforeFirstInsn) {
      if (startLabel != null) {
        // need to do this after any other handlers!
        mv.visitTryCatchBlock(startLabel, endLabel, endLabel, null);
        mv.visitLabel(startLabel);
      }
    }
    beforeFirstInsn = false;
  }

  /**
   * timestamps one object on the top of the stack, leaving it there
   */
  protected void stampDup () {
    mv.visitInsn(Opcodes.DUP);
    stampOne();
  }

  /**
   * timestamps the object on the top of the stack, consuming it
   */
  protected void stampOne() {
    mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1",
                       "(Ljava/lang/Object;)V");
  }

  /**
   * marks a value as stamped if a condition is true
   * @param val an int giving the value number to stamp conditionally
   * @param stamped a boolean indicating whether to mark the value as stamped
   */
  private void stampedIf (int val, boolean stamped) {
    if (stamped) {
      nowStamped(val);
    }
  }

  /**
   * marks a value as stamped, provided it is not an uninitialized object
   * @param val an int giving the value number to mark as stamped
   */
  private void nowStamped (int val) {
    if (val > 0) {
      this.stamped.add(val);
    }
  }

  /**
   * indicates whether we should stamp (instrument) the ith value down from the top
   * @param i an int indicating the desired number of slots below the top
   * @return a boolean that is true if we should instrument
   */
  private boolean shouldStamp (int i) {
    int vn = stack.top(i);
    return instrument && (vn > 0) && !stamped.contains(vn);
  }

  /**
   * handles optimizer tracking adjustments for operations that "tick" the stamp clock;
   * notably, it clears the set of recently stamped things so that they may be stamped
   * again with the new clock value
   */
  private void tick () {
    stamped.clear();
  }

  /**
   * found the initializing call for the given value number;
   * flip other copies to a regular value number
   * @param vn an int giving the (negative) value number being initialized
   */
  private void foundInit (int vn) {
    int newNum = ++vnCounter;
    locals.renumber(vn, newNum);
    stack .renumber(vn, newNum);
  }

  /*
   * Returns true if fieldDesc is the descriptor fora field that might point to an object.
   */
  private boolean isObjectReference (String fieldDesc) {
    char kind = fieldDesc.charAt(0);
    return (kind == 'L' || kind == '[');
  }

  /**
   * does this designate a wide (long or double) type?
   * @param desc a String giving the type descriptor
   * @return a boolean that is true iff the type is long or double
   */
  private boolean isWide (String desc) {
    char kind = desc.charAt(0);
    return (kind == 'J' || kind == 'D');
  }

  /**
   * the width of this type: 2 for long or double, 1 for others
   * @param desc a String giving the type descriptor
   * @return an int giving the type's width in slots (1 or 2)
   */
  private int width (String desc) {
    return (isWide(desc) ? 2 : 1);
  }

  /**
   * implements a subset of List for ints, without creating Integer objects
   */
  protected class IntStack {

    /**
     * number of items currently in the list
     */
    private int size;

    /**
     * holds the items
     */
    private int[] ints;

    /**
     * default initial capacity
     */
    private static final int DEFAULT_CAPACITY = 4;

    /**
     * construct an IntStack with default initial capacity
     */
    public IntStack () {
      this(DEFAULT_CAPACITY);
    }
    
    /**
     * construct an IntStack with given initial capacity
     * @param init an int giving the initial capacity
     */
    public IntStack (int init) {
      size = 0;
      ints = new int[(init >= 0) ? init : DEFAULT_CAPACITY];
    }
    
    /**
     * push an int
     * @param value an int to push
     * @return the int pushed, as a convenience
     */
    public int push (int value) {
      if (size >= ints.length) {
        int newCap = ints.length << 1;
        if (newCap > 1024) {
          newCap = ints.length + 1024;
        }
        ints = Arrays.copyOf(ints, newCap);
      }
      return (ints[size++] = value);
    }

    /**
     * push some zeroes on the sack
     * @param n an int indicating how many zeroes to push
     */
    public void pushZero (int n) {
      while (--n >= 0) {
        stack.push(0);
      }
    }

    /**
     * pop and return a value at the end
     * @return an int giving the value popped
     */
    public int pop () {
      if (size <= 0) return 0;
      return ints[--size];
    }

    /**
     * discard num items
     * @param num an int giving the number of items to discard (pop)
     */
    public void pop (int num) {
      size = ((size >= num) ? size-num : 0);
    }

    /**
     * clear the stack (make it have no items pushed)
     */
    public void clear () {
      size = 0;
    }

    /**
     * the current number of ints on the stack
     * @return an int giving the number of ints on the stack
     */
    public int size () {
      return size;
    }

    /**
     * the ith value pushed (first is numbered 0)
     * @param an int giving the index of the item desired
     * @return the int value at that index
     */
    public int get (int i) {
      if (i < 0 || i >= size)
        return 0;
      return ints[i];
    }

    /**
     * set the ith value pushed (first is numbered 0)
     * @param i an int giving the index of the item to set
     * @param v an int giving the value to store
     */
    public void set (int i, int v) {
      if (i >= 0 && i < size) {
        ints[i] = v;
      }
    }

    /**
     * the topmost value
     * @return the int on top of the stack
     */
    public int top () {
      return (size <= 0 ? 0 : ints[size-1]);
    }

    /**
     * the ith down from the top (top(0) gives topmost)
     * @param i an int indicating the number of slots down from the top desired
     * @return the int stored i slots down from the top (i == 0 gives topmost)
     */
    public int top (int i) {
      return (i < 0 || i >= size) ? 0 : ints[size-i-1];
    }

    /**
     * renumber from one value number to another
     * @param oldNum an int giving the number to be replaced wherever it occurs
     * @param newNum an int giving the new number to replace it with
     */
    public void renumber (int oldNum, int newNum) {
      for (int i = size; --i >= 0; ) {
        if (ints[i] == oldNum) {
          ints[i] = newNum;
        }
      }
    }

  }

}

// Local Variables:
// mode:Java
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
