#include "main.h"
#include "ExternalClassInstrumenter.h"
#include "ClassInstrumenter.h"
#include "CallBackHandler.h"
#include "ETCallBackHandler.h"
#include "FlatTraceOutputter.h"
#include <iostream>
#include <fstream>
#include <jni.h>
#include <jvmti.h>
#include <agent_util.h>
#include <java_crw_demo.h>
using namespace std;

unsigned char wantRecord[256];
bool includeDeathChunks = true;

void setCapabilities(jvmtiEnv *env);
void setEvents(jvmtiEnv *jvmti);
bool wantToInstrument(const char* name, jclass klass,
                      const unsigned char* klass_data, jint klass_data_len );
map<string,string> getOptions(char* optString);

static jrawMonitorID theBigLock;
static ofstream outputFile;
static jvmtiEnv* cachedJvmti;
static bool engageCallbacks = true;

bool verbose = false;
void verbose_println (string s) {
  if (verbose) {
    cout << s << endl;
  }
}

static void enter_critical_section (jvmtiEnv *jvmti) {
  jvmtiError error;

  error = jvmti->RawMonitorEnter(theBigLock);
  check_jvmti_error(jvmti, error, "Cannot enter with raw monitor");
}

/* Exit a critical section by doing a JVMTI Raw Monitor Exit */
static void exit_critical_section (jvmtiEnv *jvmti) {
  jvmtiError error;

  error = jvmti->RawMonitorExit(theBigLock);
  check_jvmti_error(jvmti, error, "Cannot exit with raw monitor");
}

void setCallbacks(jvmtiEnv *jvmti);
static ClassInstrumenter* instrumenter;
static bool vmInitialized = false;

void instrumentClass (jvmtiEnv *jvmti,
                      JNIEnv* env,
                      jclass class_being_redefined,
                      jobject loader,
                      const char* name,
                      jobject protection_domain,
                      jint class_data_len,
                      const unsigned char* class_data,
                      jint* new_class_data_len,
                      unsigned char** new_class_data) {
  if (wantToInstrument(name, class_being_redefined, class_data, class_data_len)) {
		
    /*Just pass on the call to the instrumenter */
    instrumenter->instrumentClass(jvmti, env, class_being_redefined, loader, name,
                                  protection_domain, class_data_len, class_data,
                                  new_class_data_len, new_class_data);
  }
}

static CallBackHandler* cbHandler;

static void traceBufferFull (void) {
  verbose_println("TraceBufferFull got called.");
  cbHandler->fakeAGC(cachedJvmti);
}

static void cbGCStart (jvmtiEnv *jvmti) {
  // cbHandler->GCStart(jvmti);
}

static void cbGCEnd (jvmtiEnv *jvmti) {
  cbHandler->GCEnd(jvmti);
}

static void nativePointerMovedOff (JNIEnv *env, jclass klass,
                                   jobject oldTarget, jobject origin, jobject newTarget) {
  cbHandler->nativePointerMovedOff(env, klass, oldTarget, origin, newTarget);
}

static void nativeNewObj (JNIEnv *jni, jclass klass, jobject thread, jobject o) {
  bool isWeak = isWeakRefClass(cachedJvmti, jni, jni->GetObjectClass(o));
  cbHandler->nativeNewObj(jni, klass, thread, o, isWeak);
}

static void nativeNewObj2 (JNIEnv *jni, jclass klass, jobject thread, jint classId, jint allocSite) {
  verbose_println("main::nativeNewObj2 called");
  cbHandler->nativeNewObj(jni, klass, thread, classId, allocSite, false);
}

static void nativeNewRefObj (JNIEnv *jni, jclass klass, jobject thread, jobject o) {
  //TODO: not even sure this call can be made, since newobj(Object) calls originate
  //from Object.<init>, which cannot easily tell weak ref objects from others
  cbHandler->nativeNewObj(jni, klass, thread, o, true);
}

static void nativeNewRefObj2 (JNIEnv *jni, jclass klass,
                              jobject thread, jint classId, jint allocSite) {
  cbHandler->nativeNewObj(jni, klass, thread, classId, allocSite, true);
}

static void nativeNewArray (JNIEnv *jni, jclass klass, jobject thread, jobject o) {
  cbHandler->nativeNewObj(jni, klass, thread, o, false);
}

static void nativeUninitPutfield (JNIEnv *jni, jclass klass,
                                  jobject newTarget, jint allocSite, jint fieldId) {
  cbHandler->nativeUninitPutfield(jni, klass, newTarget, allocSite, fieldId);
}

static void nativeInitOf (JNIEnv *jni, jclass klass, jobject thread, jint allocSite) {
  cbHandler->nativeInitOf(jni, klass, thread, allocSite);
}

static void nativeArrayStore (JNIEnv *jni, jclass klass,
                              jobject array, jint index, jobject storee) {
  cbHandler->nativeArrayStore(jni, klass, array, index, storee);
}

static bool isWeakRefClass1 (jvmtiEnv *jvmti_env, JNIEnv *jni_env, jclass klass) {
  char *klassSig;

  jvmtiError err = jvmti_env->GetClassSignature(klass, &klassSig, NULL);
  check_jvmti_error(jvmti_env, err,
                    "main::isWeakRefClass1: could not get class signature.");
  bool result = ((strcmp(klassSig, "Ljava/lang/ref/WeakReference;") == 0) ||
                 (strcmp(klassSig, "Ljava/lang/ref/SoftReference;") == 0));

  // We insure that the klassSig string is always freed
  jvmti_env->Deallocate((unsigned char *)klassSig);
  return result;
}

bool isWeakRefClass (jvmtiEnv *jvmti_env, JNIEnv *jni_env, jclass klass) {

  if (isWeakRefClass1(jvmti_env, jni_env, klass)) {
    return true;
  }

  klass = jni_env->GetSuperclass(klass);

  while (klass != NULL) {
    // We are careful to DeleteLocalRef all *superclasses* fetched;
    // wo don't DeleteLocalRef the original klass, since other code might use it
    if (isWeakRefClass1(jvmti_env, jni_env, klass)) {
      jni_env->DeleteLocalRef(klass);
      return true;
    }
    jclass super = jni_env->GetSuperclass(klass);
    jni_env->DeleteLocalRef(klass);
    klass = super;
  }
  return false;
}

static void cbVMAlloc (jvmtiEnv *jvmti_env,
                       JNIEnv* jni_env,
                       jthread thread,
                       jobject o,
                       jclass object_klass,
                       jlong size) {
  cbHandler->nativeNewObj(jni_env, object_klass, thread, o,
                          isWeakRefClass(jvmti_env, jni_env, object_klass));
}

void nativeMethodEntry (JNIEnv *jni, jclass klass,
                        jint methodId, jobject receiver) {
  if (vmInitialized) {
    cbHandler->nativeMethodEntry(jni, klass, methodId, receiver);
  }
}

void nativeMethodEntryStatic (JNIEnv *jni, jclass klass,
                              jint methodId) {
  if (vmInitialized) {
    cbHandler->nativeMethodEntryStatic(jni, klass, methodId);
  }
}

void nativeMethodExit (JNIEnv *jni, jclass klass,
                       jint methodId, jobject receiver) {
  if (vmInitialized) {
    cbHandler->nativeMethodExit(jni, klass, methodId, receiver);
  }
}

void nativeMethodException (JNIEnv *jni, jclass klass,
                            jobject exception, jint methodId, jobject receiver) {
  if (vmInitialized) {
    cbHandler->nativeMethodException(jni, klass, exception, methodId, receiver);
  }
}

void nativeMethodExceptionStatic (JNIEnv *jni, jclass klass,
                                  jobject exception, jint methodId) {
  if (vmInitialized) {
    cbHandler->nativeMethodExceptionStatic(jni, klass, exception, methodId);
  }
}

void nativeVarEvent (JNIEnv *jni, jclass klass, jint numObjs,
                     jobject o0, jobject o1, jobject o2, jobject o3, jobject o4) {
  if (vmInitialized) {
    cbHandler->nativeVarEvent(jni, klass, numObjs, o0, o1, o2, o3, o4);
  }
}


static void cbVMStart (jvmtiEnv *jvmti, JNIEnv *jni) {
  JNINativeMethod jni_newObj;
  JNINativeMethod jni_newObj2;
  JNINativeMethod jni_newRefObj;
  JNINativeMethod jni_newRefObj2;
  JNINativeMethod jni_newArray;
  JNINativeMethod jni_pointerMovedOff;
  JNINativeMethod jni_uninitPutfield;
  JNINativeMethod jni_initOf;
  JNINativeMethod jni_methodEntry;
  JNINativeMethod jni_methodEntryStatic;
  JNINativeMethod jni_methodExit;
  JNINativeMethod jni_methodException;
  JNINativeMethod jni_methodExceptionStatic;
  JNINativeMethod jni_arrayStore;
  JNINativeMethod jni_varEvent;
  JNINativeMethod native_methods[15];
  char* ElephantTracks_class = (char*)"ElephantTracks";
  char* ElephantTracks_engaged = (char*)"engaged";
  jclass   klass;
  int      rc;
  jfieldID engagedField;

  jni_newObj.name = (char*)"_newobj";
  jni_newObj.signature = (char*)"(Ljava/lang/Object;Ljava/lang/Object;)V";
  jni_newObj.fnPtr = (void*)nativeNewObj;
    
  jni_newObj2.name = (char*)"_newobj";
  jni_newObj2.signature = (char*)"(Ljava/lang/Object;II)V";
  jni_newObj2.fnPtr = (void*)nativeNewObj2;
    
  jni_newRefObj.name = (char*)"_newrefobj";
  jni_newRefObj.signature = (char*)"(Ljava/lang/Object;Ljava/lang/Object;)V";
  jni_newRefObj.fnPtr = (void*)nativeNewRefObj;
    
  jni_newRefObj2.name = (char*)"_newrefobj";
  jni_newRefObj2.signature = (char*)"(Ljava/lang/Object;II)V";
  jni_newRefObj2.fnPtr = (void*)nativeNewRefObj2;
    
  jni_newArray.name = (char*)"_newarray";
  jni_newArray.signature = (char*)"(Ljava/lang/Object;Ljava/lang/Object;)V";
  jni_newArray.fnPtr = (void*)nativeNewArray;
    
  jni_pointerMovedOff.name = (char*)"_pointerMovedOff";
  jni_pointerMovedOff.signature = (char*)"(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V";
  jni_pointerMovedOff.fnPtr = (void*)nativePointerMovedOff;
    
  jni_uninitPutfield.name = (char*)"_uninitPutfield";
  jni_uninitPutfield.signature = (char*)"(Ljava/lang/Object;II)V";
  jni_uninitPutfield.fnPtr = (void*)nativeUninitPutfield;
    
  jni_initOf.name = (char*)"_initOf";
  jni_initOf.signature = (char*)"(Ljava/lang/Object;I)V";
  jni_initOf.fnPtr = (void*)nativeInitOf;
    
  jni_methodEntry.name = (char*)"_methodEntry";
  jni_methodEntry.signature = (char*)"(ILjava/lang/Object;)V";
  jni_methodEntry.fnPtr = (void*)nativeMethodEntry;
    
  jni_methodEntryStatic.name = (char*)"_methodEntryStatic";
  jni_methodEntryStatic.signature = (char*)"(I)V";
  jni_methodEntryStatic.fnPtr = (void*)nativeMethodEntryStatic;
    
  jni_methodExit.name = (char*)"_methodExit";
  jni_methodExit.signature = (char*)"(ILjava/lang/Object;)V";
  jni_methodExit.fnPtr = (void*)nativeMethodExit;
   
  jni_methodException.name = (char*)"_methodException";
  jni_methodException.signature = (char*)"(Ljava/lang/Object;ILjava/lang/Object;)V";
  jni_methodException.fnPtr = (void*)nativeMethodException;
   
  jni_methodExceptionStatic.name = (char*)"_methodExceptionStatic";
  jni_methodExceptionStatic.signature = (char*)"(Ljava/lang/Object;I)V";
  jni_methodExceptionStatic.fnPtr = (void*)nativeMethodExceptionStatic;
   
  jni_arrayStore.name = (char*)"_arrayStore";
  jni_arrayStore.signature = (char*)"(Ljava/lang/Object;ILjava/lang/Object;)V";
  jni_arrayStore.fnPtr = (void*)nativeArrayStore;

  jni_varEvent.name = (char*)"_varEvent";
  jni_varEvent.signature = (char*)
    "(ILjava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V";
  jni_varEvent.fnPtr = (void*)nativeVarEvent;
    
  klass = jni->FindClass(ElephantTracks_class);                                         
    
  if (klass == NULL) {
    fatal_error("ERROR: JNI: Cannot find %s with FindClass.\n", ElephantTracks_class);
  }
    
  native_methods[ 0] = jni_newObj;
  native_methods[ 1] = jni_newObj2;
  native_methods[ 2] = jni_newRefObj;
  native_methods[ 3] = jni_newRefObj2;
  native_methods[ 4] = jni_newArray;
  native_methods[ 5] = jni_methodEntry;
  native_methods[ 6] = jni_methodEntryStatic;
  native_methods[ 7] = jni_methodExit;
  native_methods[ 8] = jni_methodException;
  native_methods[ 9] = jni_methodExceptionStatic;
  native_methods[10] = jni_arrayStore;
  native_methods[11] = jni_pointerMovedOff;
  native_methods[12] = jni_uninitPutfield;
  native_methods[13] = jni_initOf;
  native_methods[14] = jni_varEvent;
    
  rc = jni->RegisterNatives(klass, native_methods, 15);
  if (rc != 0) {
    fatal_error("ElephantTracksAgent: ERROR: JNI: Cannot register native methods for %s, rc: %d\n",
                ElephantTracks_class, rc);
  }
   
  if (engageCallbacks) {
    /* Engage calls. */ 
    engagedField = jni->GetStaticFieldID(klass, ElephantTracks_engaged, "I");
    if (engagedField == NULL) {
      fatal_error("ERROR: JNI: Cannot get field from %s\n", ElephantTracks_class);
    }
    jni->SetStaticIntField(klass, engagedField, 1); 
  }
}

static void cbVMInit (jvmtiEnv *jvmti, JNIEnv *env, jthread thread) {
  cbHandler->VMInit(jvmti, env, thread);
  vmInitialized = true;
}

jint Agent_OnLoad (JavaVM *vm, char *options, void *reserved) {
  jvmtiEnv *jvmti;
  TraceOutputter *outputter;
  jvmtiError error;
  map<string,string> optionMap;

  optionMap = getOptions(options);

  //First, check if we need to be verbose
  if (optionMap["verbose"] == "T" || optionMap["verbose"] == "t") {
    verbose = true;
  }

  outputFile.open((optionMap["traceFile"]).c_str());

  if (outputFile.fail()) {
    cerr << "Error opening output file " << optionMap["traceFile"] << " for writing." << endl;
    exit(1);
  }

  unsigned int maxRecordsToBuffer = atoi(optionMap["bufferSize"].c_str()) ;

  string which = optionMap["whichRecs"];
  int whichDefault = (which[0] == '-');
  for (int i = 0; i < 256; ++i) {
    wantRecord[i] = whichDefault;
  }
  int whichSet = 1;
  for (int i = 0; i < which.length(); ++i) {
    if (which[i] == '-') {
      whichSet ^= 1;
    } else {
      wantRecord[which[i]] = whichSet;
    }
  }

  if (optionMap.count("engageCallbacks") && optionMap["engageCallbacks"].compare("false") == 0) {
    engageCallbacks = false;
  }

  if(optionMap.count("deathChunks") && optionMap["deathChunks"].compare("false") == 0) {
    includeDeathChunks = false;
  }

  vm->GetEnv((void**)&jvmti,JVMTI_VERSION_1);

  instrumenter =
    new ExternalClassInstrumenter(jvmti, optionMap["javaPath"], optionMap["classPath"],
                                  optionMap["classReWriter"], optionMap["useOptimizer"],
                                  optionMap["namesFile"], verbose, optionMap["experimental"]);
  outputter = new FlatTraceOutputter(jvmti, &outputFile, maxRecordsToBuffer, &traceBufferFull);
  //outputter = new TraceOutputter();
  cbHandler = new ETCallBackHandler(jvmti, vm, outputter);
  //cbHandler = new DCCallBackHandler(jvmti, outputter);	

  cachedJvmti = jvmti;

  /* Here we create a raw monitor for our use in this agent to
   *   protect critical sections of code.
   */
  error = jvmti->CreateRawMonitor("agent data", &theBigLock);
  check_jvmti_error(jvmti, error, "Cannot create raw monitor");
	
  setCapabilities(jvmti);
  setEvents(jvmti);	
  setCallbacks(jvmti);

  return JNI_OK;	
}


static void cbVMDeath (jvmtiEnv *jvmti, JNIEnv *env) {
  jvmtiError          error;
  jclass              klass;
  jfieldID            field;
  jvmtiEventCallbacks callbacks;
  const char* ElephantTracks_class = "ElephantTracks";
  const char* ElephantTracks_engaged = "engaged";
        
  vmInitialized = false;
		
  klass = env->FindClass(ElephantTracks_class);
  if (klass == NULL) {
    fatal_error("ERROR: JNI: Cannot find %s with FindClass\n", ElephantTracks_class);
  }

  field = env->GetStaticFieldID(klass, ElephantTracks_engaged, "I");
  if (field == NULL) {
    fatal_error("ERROR: JNI: Cannot get field from %s\n", ElephantTracks_class);
  }
  env->SetStaticIntField(klass, field, 0);
		 
  /* Clear out all callbacks. */
  (void)memset(&callbacks, 0, sizeof(callbacks));
  error = jvmti->SetEventCallbacks(&callbacks, (jint)sizeof(callbacks));
  check_jvmti_error(jvmti, error, "Cannot set jvmti callbacks");

  cbHandler->VMDeath(jvmti);

  outputFile.close();
  delete instrumenter;

  cerr << "Elephant Tracks Completed." << endl;
}   

void setCallbacks (jvmtiEnv *jvmti) {
  jvmtiEventCallbacks callbacks;
  jvmtiError error;	

  (void)memset(&callbacks, 0, sizeof(callbacks));
  callbacks.ClassFileLoadHook = &(instrumentClass);
  callbacks.VMStart = &(cbVMStart);
  callbacks.GarbageCollectionStart = &(cbGCStart);
  callbacks.GarbageCollectionFinish = &(cbGCEnd);
  callbacks.VMInit = &(cbVMInit);
  callbacks.VMDeath = &(cbVMDeath);
  callbacks.VMObjectAlloc = &(cbVMAlloc);
	
  error = jvmti->SetEventCallbacks(&callbacks, (jint)sizeof(callbacks));
  check_jvmti_error(jvmti, error, "Cannot set jvmti callbacks");	
}

void setCapabilities (jvmtiEnv *env) {
  jvmtiCapabilities capa;
  jvmtiError error;
	
  (void)memset(&capa, 0, sizeof(capa));
     
  /*Tell the JVM what capabilities we want*/
  capa.can_generate_all_class_hook_events  = 1;
  capa.can_generate_frame_pop_events = 1;

  capa.can_tag_objects = 1;
  capa.can_generate_vm_object_alloc_events = 1;
  capa.can_generate_garbage_collection_events = 1;

  //Don't need method entry/exit any more
  //capa.can_generate_method_entry_events = 1;
  //capa.can_generate_method_exit_events = 1;
  //Don't need local var access eiter...
  //capa.can_access_local_variables = 1;

  capa.can_generate_garbage_collection_events = 1;

  error = env->AddCapabilities(&capa);
  check_jvmti_error(env, error, "Unable to get necessary JVMTI capabilities.");
}

void setEvents (jvmtiEnv *jvmti) {
  jvmtiError error;
		 
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE,
                                          JVMTI_EVENT_CLASS_FILE_LOAD_HOOK, (jthread)NULL);

  check_jvmti_error(jvmti, error,
                    "Cannot set event notification JVMTI_EVENT_CLASS_FILE_LOAD_HOOK");
         
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_VM_START, (jthread)NULL);
  check_jvmti_error(jvmti, error, "Cannot set event notification JVMTI_EVENT_VM_START");
	 
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_GARBAGE_COLLECTION_START,
                                          (jthread)NULL);
  check_jvmti_error(jvmti, error,
                    "Cannot set event notification JVMTI_EVENT_GARBAGE_COLLECTION_START");
	 
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_GARBAGE_COLLECTION_FINISH,
                                          (jthread)NULL);
  check_jvmti_error(jvmti, error,
                    "Cannot set event notification JVMTI_EVENT_GARBAGE_COLLECTION_FINISH");
	 
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_VM_INIT, (jthread)NULL);
  check_jvmti_error(jvmti, error, "Cannot set event notification JVMTI_EVENT_VM_INIT"); 
	 
  error = jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_VM_DEATH, (jthread)NULL);
  check_jvmti_error(jvmti, error, "Cannot set event notification JVMTI_EVENT_VM_DEATH");
	 
  error= jvmti->SetEventNotificationMode(JVMTI_ENABLE, JVMTI_EVENT_VM_OBJECT_ALLOC, (jthread)NULL);
  check_jvmti_error(jvmti, error, "Cannot set event notification JVMTI_EVENT_VM_OBJECT_ALLOC"); 
}

bool wantToInstrument (const char* name, jclass klass,
                       const unsigned char* klass_data, jint klass_data_len) {
  const char *classname;

  if (name == NULL) {/*Then we have to go fishing for the name in the class itself */	          
    classname = java_crw_demo_classname(klass_data, klass_data_len, NULL);

    if (classname == NULL) {
      fatal_error("ERROR: No classname inside classfile\n");
    }
  } else {
    /*else {
      classname = strdup(name);
      if ( classname == NULL ) {
      fatal_error("ERROR: Out of malloc memory\n");
      }
      }*/
    classname = name;
  }

  if (string(classname) == "ElephantTracks") {
    return false;
  } else {
    return true;
  }
		
}

pair<string,string> parseOption (string option) {
  string opt, val;
  for (unsigned int i = 0; i < option.size(); i++) {
    if (option[i] == '=') {
      opt = option.substr(0,i);
      val = option.substr(i+1, option.size()-1);
    }
  }

  return pair<string,string>(opt,val);
}

map<string,string> parseOptionList (string options) {
  map<string,string> opt2Val;
  pair<string,string> optionVal;
  unsigned int beginning = 0;
  char sep = ':';

  // support :=c at start of options, to change the separator;
  // useful if an option contains a ':'
  if (options.length() > 2 && options[0] == ':' && options[1] == '=') {
    sep = options[2];
    beginning = 3;
  }

  for (unsigned int i = beginning; i < options.length(); i++) {
    if (options[i] == sep) {
      opt2Val.insert(parseOption(options.substr(beginning, i-beginning)));
      beginning = i+1;
    }
  }

  //handle the last option
  if (beginning < options.length()) {
    opt2Val.insert(parseOption(options.substr(beginning, options.length()-1)));
  }

  return opt2Val;
}

map<string,string> getOptions (char* optString) {
  map<string,string> optMap;

  optMap = parseOptionList(string(optString));

  //Set some defaults if they were not set
  if (optMap.find("javaPath") == optMap.end()) {
    optMap["javaPath"] = "/usr/bin/java";
  }
  if (optMap.find("classPath") == optMap.end()) {
    optMap["classPath"] = "/h/sguyer/Projects/KeyObjects/asm-3.1.jar";
  }
  if (optMap.find("outputFile") == optMap.end()) {
    optMap["outputFile"] = "elephantTracksOut";
  }
  if (optMap.find("traceFile") == optMap.end()) {
    optMap["traceFile"] = optMap["outputFile"] + ".trace";
  }
  if (optMap.find("namesFile") == optMap.end()) {
    optMap["namesFile"] = optMap["outputFile"] + ".names";
  }
  if (optMap.find("whichRecs") == optMap.end()) {
    optMap["whichRecs"] = "ADMEUNR";
  }
  if (optMap.find("bufferSize") == optMap.end()) {
    optMap["bufferSize"] =  "10000000";
  }
  if (optMap.find("verbose") == optMap.end()) {
    optMap["verbose"] = "F";
  }
  if (optMap.find("experimental") == optMap.end()) {
    optMap["experimental"] = "";
  } else {
    optMap["experimental"] = "-x" + optMap["experimental"];
  }
  if (optMap.find("useOptimizer") == optMap.end()) {
    optMap["useOptimizer"] = "";
  } else {
    const char *useOpt = optMap["useOptimizer"].c_str();
    if (useOpt[0] == 't' || useOpt[0] == 'T') {
      optMap["useOptimizer"] = "-opt";
    } else {
      optMap["useOptimizer"] = "";
    }
  }

  return optMap;
}

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
